/********************************************************************************
* Fast and Exact Gate Sizing by Lagrangian Relaxation                           *
* File name: Gate.cpp: implementation of the CGate class                        *
* Author: Yu-Hao Wang							        *
* Created      : 09/28/01							*
* Last modified: 02/01/02							*
*********************************************************************************/

#include "gate.h"

CGate::CGate()
{
  strcpy(m_name, "");
  m_id		   = 0;
  m_nFanIn         = 0;
  m_nFanOut        = 0;
  m_vArrivalTime.clear();
  m_Delay          = 0.0;
  m_r              = 0.0;
  m_R              = 0.0;
  m_c              = 0.0;
  m_C              = 0.0;
  m_cFET           = 0.0;
  m_mu             = 0.0;
  m_vLambda.clear();
  m_size           = 0.0;
  m_sizeBound[0]   = 16.9*Femto; 
  m_sizeBound[1]   = 10*Nano;
  m_primaryOutput  = false;
}

CGate::CGate( const CGate & One )
{
  strcpy(m_name, One.m_name);
  m_id		     = One.m_id;
  m_nFanIn           = One.m_nFanIn;
  m_nFanOut          = One.m_nFanOut;
  m_vArrivalTime     = One.m_vArrivalTime;
  m_Delay            = One.m_Delay;
  m_r                = One.m_r;
  m_R                = One.m_R;
  m_c                = One.m_c;
  m_C                = One.m_C;
  m_cFET             = One.m_cFET;
  m_mu               = One.m_mu;
  m_vLambda          = One.m_vLambda;
  m_size             = One.m_size;
  m_sizeBound[0]     = One.m_sizeBound[0];
  m_sizeBound[1]     = One.m_sizeBound[1];
  m_primaryOutput    = One.m_primaryOutput;
}

CGate::CGate(const char* name, int id)
{
  strcpy(m_name, name);
  m_id             = id;
  m_nFanIn         = 0;
  m_nFanOut        = 0;
  m_vArrivalTime.clear();
  m_Delay          = 0.0;
  m_r              = 0.0;
  m_R              = 0.0;
  m_c              = 0.0;
  m_C              = 0.0;
  m_cFET           = 0.0;
  m_mu             = 0.0;
  m_vLambda.clear();
  m_size           = 0.0;
  m_sizeBound[0]   = 16.9*Femto; 
  m_sizeBound[1]   = 10*Nano; 
  m_primaryOutput  = false;
}

const CGate& CGate::operator=(const CGate & One)
{
  if ( &One != this )
    {
      strcpy(m_name, One.m_name);
      m_id	      = One.m_id;
      m_nFanIn        = One.m_nFanIn;
      m_nFanOut       = One.m_nFanOut;
      m_vArrivalTime  = One.m_vArrivalTime;
      m_Delay         = One.m_Delay;
      m_r             = One.m_r;
      m_R             = One.m_R;
      m_c             = One.m_c;
      m_C             = One.m_C;
      m_cFET          = One.m_cFET;
      m_mu            = One.m_mu;
      m_vLambda       = One.m_vLambda;
      m_size          = One.m_size;
      m_sizeBound[0]  = One.m_sizeBound[0];
      m_sizeBound[1]  = One.m_sizeBound[1];
      m_primaryOutput = One.m_primaryOutput;
    }
  return *this;
}

bool CGate::operator==(const CGate & One) const
{
  return (strcmp(m_name, One.m_name) == 0 &&
	  m_id == One.m_id);
}

bool CGate::operator<( const CGate &One ) const
{
	return m_id < One.m_id;
}

CGate::~CGate()
{

}
