/********************************************************************************
* Fast and Exact Gate Sizing by Lagrangian Relaxation                           *
* File name: Netlist.cpp; implementation of the CNetlist class                  *
* Author: Yu-Hao Wang				   				*
* Created      : 09/28/01							*
* Last modified: 02/01/02					  		*
*********************************************************************************/

#include "netlist.h"

CNetlist::CNetlist()
{
  m_size         = 0;
  m_nGate        = 0;
  m_nWire        = 0;
  m_nPris        = 0;
  m_nPos         = 0;
  m_nLine        = 0;
  m_Condition[0] = 0.005;
  m_Condition[1] = 0.01;
  m_driverR      = 5000;
  m_loadC        = 2.3995*Femto;
  m_Total_Cost   = 0.0;
  m_Total_Delay  = 0.0;
  m_TO.clear();
  m_RTO.clear();
  m_PartArray.clear();
  m_GateList.clear();
}

CNetlist::~CNetlist()
{
}

double CNetlist::maximum(vector<double> One)
{
  if (One.size() > 0) {
    sort(One.begin(), One.end());
    return One.back();
  }
  else
    return 0.0;
}

void CNetlist::BuildDB(char* filename)
{
  FILE *stream;
  char line[256];
  char seps[] = "\t,\n";
  char *token;
  CGateType g;
  CGateData data;
  list<CGateType>::iterator p1;

  cout << "Running QuickTune ";

  m_beginTime = clock();
  time( &m_time );

  if( ( stream = fopen(filename, "r") ) != NULL )
    {
      if (fgets( line, 256, stream ) != NULL)
	{
	  while( 1 ) {

	    if (fgets( line, 256, stream ) != NULL)
	    {

	      token = strtok( line, seps );
	      if (token != NULL) {
		strcpy(g.m_key, token);
	      }
	      else {
		cerr << "Format Error: " << filename << " could not be parsed" << endl;
		exit( 1 );
	      }

	      token = strtok( NULL, seps );
	      if (token != NULL)
		data.m_key = atoi(token);
	      else {
		cerr << "Format Error: " << filename << " could not be parsed" << endl;
		exit( 1 );
	      }

	      token = strtok( NULL, seps );
	      if (token != NULL)
		data.m_c = atof(token);
	      else {
		cerr << "Format Error: " << filename << " could not be parsed" << endl;
		exit( 1 );
	      }

	      token = strtok( NULL, seps );
	      if (token != NULL)
		data.m_r = atof(token);
	      else {
		cerr << "Format Error: " << filename << " could not be parsed" << endl;
		exit( 1 );
	      }

	      token = strtok( NULL, seps );
	      if (token != NULL)
		data.m_cFET = atof(token);
	      else {
		cerr << "Format Error: " << filename << " could not be parsed" << endl;
		exit( 1 );
	      }

	      token = strtok( NULL, seps );
	      if (token != NULL)
		data.m_alpha = atof(token);
	      else {
		cerr << "Format Error: " << filename << " could not be parsed" << endl;
		exit( 1 );
	      }

	      token = strtok( NULL, seps );
	      if (token != NULL)
		data.m_beta = atof(token);
	      else {
		cerr << "Format Error: " << filename << " could not be parsed" << endl;
		exit( 1 );
	      }

	    }
	    else
	      break;

	    p1 = find(m_GateList.begin(), m_GateList.end(), g);
	    if (p1 == m_GateList.end()) {
	      g.m_dataArray.clear();
	      g.m_dataArray.push_back(data);
	      m_GateList.push_back(g);
	    }
	    else
	      p1->m_dataArray.push_back(data);
	  }

	  for (p1 = m_GateList.begin(); p1 != m_GateList.end(); ++p1)
	    sort(p1->m_dataArray.begin(), p1->m_dataArray.end());
	  m_GateList.sort();

	}
      else {
	cerr << "Format Error: " << filename << " could not be parsed" << endl;
	exit( 1 );
      }

      fclose( stream );
    }
  else {
    cerr << "File Error: " << filename << " could not be opened" << endl;
    exit( 1 );
  }

}

void CNetlist::InputNetlist(char* filename)
{
  FILE *stream;
  int i;
  char line[256];
  char seps[] = " ,\n";
  char *token;
  char String[10];
  CGateType g;
  vector<PartType>::iterator p1;
  list<PartInfo>::iterator p2;
  list<CGateType>::iterator p3;

  if( ( stream = fopen(filename, "r") ) != NULL )
    {
      if ( fgets( line, 256, stream ) != NULL )
	sscanf( line, "%d %d %d %d %d", &m_nGate, &m_nWire, &m_nPris, &m_nPos, &m_nLine );
      else {
	cerr << "Format Error: " << filename << " could not be parsed" << endl;
	exit( 1 );
      }

      for (i=1; i<=m_nLine; i++) 
	{
	  if ( fgets( line, 256, stream ) != NULL ) 
	    {
	      PartType p;
	      PartInfo pInfo;
	      p.part.m_id = i;
	      
	      token = strtok( line, seps );
	      if (token != NULL)
		strcpy(String, token);
	      else { 
		cerr << "Format Error: " << filename << " could not be parsed" << endl;
		exit( 1 );
	      }

	      if (strcmp(String, "po") != 0) {
		strcpy(p.part.m_name, String);

		token = strtok( NULL, seps );
		if (token != NULL)
		  p.part.m_nFanOut = atoi(token);
		else { 
		  cerr << "Format Error: " << filename << " could not be parsed" << endl;
		  exit( 1 );
		}
		token = strtok( NULL, seps );
		if (token != NULL)
		  p.part.m_nFanIn = atoi(token);
		else { 
		  cerr << "Format Error: " << filename << " could not be parsed" << endl;
		  exit( 1 );
		}
		

		if (strcmp(String, "pri") != 0) {
		  strcpy(g.m_key, p.part.m_name);
		  p3 = find(m_GateList.begin(), m_GateList.end(), g);
		  if (p3 != m_GateList.end()) {
		    if (p.part.m_nFanIn >= 1) {
		      if ( (p.part.m_nFanIn - p3->m_dataArray.front().m_key) >= 0) {

			p.part.m_c    = p3->m_dataArray[p.part.m_nFanIn - p3->m_dataArray.front().m_key].m_c;
			p.part.m_r    = p3->m_dataArray[p.part.m_nFanIn - p3->m_dataArray.front().m_key].m_r;
			p.part.m_cFET = p3->m_dataArray[p.part.m_nFanIn - p3->m_dataArray.front().m_key].m_cFET;

		      }
		      else {
			cerr << "Data Error: [" << g.m_key << "] could not be read" << endl;
			exit( 1 );
		      }
		    }
		    else {
		      cerr << "Data Error: [" << g.m_key << "] could not be read" << endl;
		      exit( 1 );
		    }
		  }
		  else {
		    cerr << "Data Error: [" << g.m_key << "] could not be read" << endl;
		    exit( 1 );
		  }
		}

	      }

	      else {
		p.part.m_primaryOutput = true;

		token = strtok( NULL, seps );
		if (token != NULL)
		  strcpy(p.part.m_name, token);
		else { 
		  cerr << "Format Error: " << filename << " could not be parsed" << endl;
		  exit( 1 );
		}

		token = strtok( NULL, seps );
		if (token != NULL)
		  p.part.m_nFanOut = atoi(token);
		else { 
		  cerr << "Format Error: " << filename << " could not be parsed" << endl;
		  exit( 1 );
		}
		
		token = strtok( NULL, seps );
		if (token != NULL)
		  p.part.m_nFanIn = atoi(token);
		else { 
		  cerr << "Format Error: " << filename << " could not be parsed" << endl;
		  exit( 1 );
		}
		
		
		strcpy(g.m_key, p.part.m_name);
		p3 = find(m_GateList.begin(), m_GateList.end(), g);
		if (p3 != m_GateList.end()) {  // Found
		  if (p.part.m_nFanIn >= 1) {
		    if ((p.part.m_nFanIn - p3->m_dataArray.front().m_key) >= 0) {
		      
		      p.part.m_c    = p3->m_dataArray[p.part.m_nFanIn - p3->m_dataArray.front().m_key].m_c;
		      p.part.m_r    = p3->m_dataArray[p.part.m_nFanIn - p3->m_dataArray.front().m_key].m_r;
		      p.part.m_cFET = p3->m_dataArray[p.part.m_nFanIn - p3->m_dataArray.front().m_key].m_cFET;

		    }
		    else { 
		      cerr << "Data Error: [" << g.m_key << "] could not be read" << endl;
		      exit( 1 );
		    }
		  }
		  else { 
		    cerr << "Data Error: [" << g.m_key << "] could not be read" << endl;
		    exit( 1 );
		  }
		}
		else { 
		  cerr << "Data Error: [" << g.m_key << "] could not be read" << endl;
		  exit( 1 );
		}
		
	      }

	      for(int j=0; j<p.part.m_nFanIn; j++) {
		
		token = strtok( NULL, seps );
		if (token != NULL)
		  pInfo.id = atoi(token);
		else { 
		  cerr << "Format Error: " << filename << " could not be parsed" << endl;
		  exit( 1 );
		}
		
		token = strtok( NULL, seps );
		if (token != NULL)
		  pInfo.rt = atof(token);
		else { 
		  cerr << "Format Error: " << filename << " could not be parsed" << endl;
		  exit( 1 );
		}
		
		token = strtok( NULL, seps );
		if (token != NULL)
		  pInfo.ft = atof(token);
		else { 
		  cerr << "Format Error: " << filename << " could not be parsed" << endl;
		  exit( 1 );
		}
		
		pInfo.pinNum = j;
		p.part.m_vLambda.push_back(-1);
		p.part.m_vArrivalTime.push_back(0.0);
		p.inputList.push_back(pInfo);
	      }
	      m_PartArray.push_back(p);
	    }
	  else {
	    cerr << "Format Error: " << filename << " could not be parsed" << endl;
	    exit( 1 );
	  }
	}
      
      fclose( stream );
      
    }
  else {
    cerr << "File Error: " << filename << " could not be opened" << endl;
    exit( 1 );
  }

  for (int k=i; k<=i+1; k++) {
    
    PartType p;
    PartInfo pInfo;
    p.part.m_id = k;

    if (k == i) {
      strcpy(p.part.m_name, "Input");
      for (p1 = m_PartArray.begin(); p1 != m_PartArray.end(); ++p1) {
	if (strcmp(p1->part.m_name, "pri") == 0) {
	  PartInfo One;
	  One.id = p.part.m_id;
	  One.pinNum = 0;
	  p1->inputList.push_back(One);
	  p1->part.m_nFanIn++;
	  p1->part.m_r = m_driverR;
	  p1->part.m_vArrivalTime.push_back(0.0);
	  p1->part.m_vLambda.push_back(-1);
	  pInfo.id = p1->part.m_id;
	  pInfo.pinNum = 0;
	  p.outputList.push_back(pInfo);
	}
      }
      p.part.m_nFanOut = p.outputList.size();
      p.part.m_nFanIn = 0;
      m_PartArray.push_back(p);
    }

    else {
      strcpy(p.part.m_name, "Output");
      p.part.m_c     = m_loadC; 
      pInfo.pinNum   = 0;
      for (p1 = m_PartArray.begin(); p1 != m_PartArray.end(); ++p1) {
	if (p1->part.m_primaryOutput == true) {
	  p1->part.m_nFanOut++;
	  pInfo.id = p1->part.m_id;
	  p.inputList.push_back(pInfo);
	  p.part.m_vLambda.push_back(-1);
	  p.part.m_vArrivalTime.push_back(0.0);
	  pInfo.pinNum++;
	}
      }
      p.part.m_nFanIn = p.inputList.size();
      p.part.m_nFanOut = 0;
      m_PartArray.push_back(p);
    }		
  }
  
  m_size = m_PartArray.size();
  
  PartInfo pInfo;

  for (p1 = m_PartArray.begin(); p1 != m_PartArray.end(); ++p1) {
    if (strcmp(p1->part.m_name, "pri") != 0 &&
	strcmp(p1->part.m_name, "Input") != 0) {
      pInfo.id     = p1->part.m_id;
      pInfo.pinNum = 0;
      
      for (p2 = (*p1).inputList.begin(); p2 != (*p1).inputList.end(); ++p2) {
	m_PartArray[p2->id - 1].outputList.push_back(pInfo);
	pInfo.pinNum++;
      }
    }
  }

  TopologicalSort(m_PartArray, m_TO, true);
  TopologicalSort(m_PartArray, m_RTO, false);

}

void CNetlist::TopologicalSort(vector<PartType> partArray, vector<int>& TopOrderArray, bool status)
{
  vector<PartType>::iterator p1;
  list<PartInfo>::iterator p2;
  vector<PartType>::iterator p3;
  vector<int>::reverse_iterator p4;
  
  PartType One;
  
  vector<int> tempArray;
  tempArray.clear();
  TopOrderArray.clear();
  
  while (!partArray.empty()) {

    for (p1 = partArray.begin(); p1 != partArray.end(); ++p1) {

      if ((*p1).part.m_nFanOut == 0) {
	for (p2 = (*p1).inputList.begin(); p2 != (*p1).inputList.end(); ++p2) {
	  One.part.m_id = p2->id;
	  p3 = find(partArray.begin(), partArray.end(), One);
	  if (p3 != partArray.end())
	    p3->part.m_nFanOut--;
	}
	
	tempArray.push_back(p1->part.m_id);

	partArray.erase(p1);
	--p1;
      }
    }
  }
  
  if (status) {
    for (p4 = tempArray.rbegin(); p4 != tempArray.rend(); ++p4)
      TopOrderArray.push_back(*p4);
  }
  else
    TopOrderArray = tempArray;
  
}

void CNetlist::ComputeArrivalTime()
{
  vector<int>::iterator p1;
  list<PartInfo>::iterator p2;

  for (p1 = m_TO.begin(); p1 != m_TO.end(); ++p1) {
    for (p2 = m_PartArray[*p1 - 1].inputList.begin(); p2 != m_PartArray[*p1 - 1].inputList.end(); ++p2) 
      m_PartArray[*p1 - 1].part.m_vArrivalTime[p2->pinNum] = maximum(m_PartArray[p2->id - 1].part.m_vArrivalTime) + m_PartArray[p2->id - 1].part.m_Delay;
  }
}

void CNetlist::IntializeLM()
{
  vector<int>::iterator p1;
  list<PartInfo>::iterator p2;
  double value = 0.0;

  for (p1 = m_RTO.begin(); p1 != m_RTO.end(); ++p1) {

    if (strcmp(m_PartArray[*p1 - 1].part.m_name,"Output") == 0) {
      value = 0.0;
      for (int i=0; i<m_PartArray[*p1 - 1].part.m_vLambda.size(); i++) {
	m_PartArray[*p1 - 1].part.m_vLambda[i] = 1 / (1.0 * m_nPos);
	value += m_PartArray[*p1 - 1].part.m_vLambda[i];
      }
      m_PartArray[*p1 - 1].part.m_mu = value;
    }
    else {  
      double Total_Lambda = 0.0;
      for (p2 = m_PartArray[*p1 - 1].outputList.begin(); p2 != m_PartArray[*p1 - 1].outputList.end(); ++p2)	
	Total_Lambda += m_PartArray[p2->id - 1].part.m_vLambda[p2->pinNum];
      
      value = 0.0;
      for (int i=0; i<m_PartArray[*p1 - 1].part.m_vLambda.size(); i++) {
	m_PartArray[*p1 - 1].part.m_vLambda[i] = Total_Lambda / (1.0 * m_PartArray[*p1 - 1].part.m_nFanIn);
	value += m_PartArray[*p1 - 1].part.m_vLambda[i];
      }
      m_PartArray[*p1 - 1].part.m_mu = value;
    }
  }
}


void CNetlist::UpdateLM()
{
  vector<int>::iterator p1;
  list<PartInfo>::iterator p2;
  char String[10];
  int i;
  double value = 0.0;

  for (p1 = m_RTO.begin(); p1 != m_RTO.end(); ++p1) 
    {
      strcpy(String, m_PartArray[*p1 - 1].part.m_name);
      
      if (strcmp(String, "Input") != 0) {
	double Output_Arrival_Time = maximum(m_PartArray[*p1 - 1].part.m_vArrivalTime) + m_PartArray[*p1 - 1].part.m_Delay;

	double Factor = 0.0;
	double Total_Lambda = 0.0;
	for (i=0; i<m_PartArray[*p1 - 1].part.m_vLambda.size(); i++) 
	  Factor += m_PartArray[*p1 - 1].part.m_vLambda[i] * (m_PartArray[*p1 - 1].part.m_vArrivalTime[i] + m_PartArray[*p1 - 1].part.m_Delay) / (1.0 * Output_Arrival_Time);
	
	if (strcmp(String, "Output") == 0) {
	  m_Total_Delay = Output_Arrival_Time;
	  Total_Lambda = 1.0;
	}
	else {
	  for (p2 = m_PartArray[*p1 - 1].outputList.begin(); p2 != m_PartArray[*p1 - 1].outputList.end(); ++p2)	
	    Total_Lambda += m_PartArray[p2->id - 1].part.m_vLambda[p2->pinNum];
	}
	
	Factor = Total_Lambda / (1.0 * Factor);

	value = 0.0;
	for (i=0; i<m_PartArray[*p1 - 1].part.m_vLambda.size(); i++) {
	  m_PartArray[*p1 - 1].part.m_vLambda[i] = m_PartArray[*p1 - 1].part.m_vLambda[i] * (m_PartArray[*p1 - 1].part.m_vArrivalTime[i] + m_PartArray[*p1 - 1].part.m_Delay) / (1.0 * Output_Arrival_Time) * Factor;
	  value += m_PartArray[*p1 - 1].part.m_vLambda[i];
	}
	m_PartArray[*p1 - 1].part.m_mu = value;	
      }
    }
}

double CNetlist::ComputeLu(vector<PartType>& One)
{
  vector<PartType>::iterator p1;
  char String[10];
  double value = 0.0;

  for (p1 = One.begin(); p1 != One.end(); ++p1) {
    strcpy(String, p1->part.m_name);

    if (strcmp(String, "Input")  != 0 &&
	strcmp(String, "Output") != 0) {

      if (strcmp(String, "pri") == 0)
	p1->part.m_Delay = (p1->part.m_r * p1->part.m_C);
      else
	p1->part.m_Delay = ((p1->part.m_r / p1->part.m_size) * p1->part.m_C);

      value += (p1->part.m_mu * p1->part.m_Delay);
    }
  }
  return value;
}

void CNetlist::SolveLRSuHelper()
{
  vector<PartType>::iterator p1;
  list<PartInfo>::iterator p2;
  vector<int>::iterator p3;
  char String[10];

  for (p1 = m_PartArray.begin(); p1 != m_PartArray.end(); ++p1) {
    strcpy(String, p1->part.m_name);
    if (strcmp(String, "pri")    != 0 &&
	strcmp(String, "Input")  != 0 &&
	strcmp(String, "Output") != 0) {
      p1->part.m_size = p1->part.m_sizeBound[0];
    }
  }

  for (p3 = m_RTO.begin(); p3 != m_RTO.end(); ++p3) {
    strcpy(String, m_PartArray[*p3 - 1].part.m_name);

    if (m_PartArray[*p3 - 1].part.m_primaryOutput == true) {

      double Downstream_Cap = 0.0;
      for (p2 = m_PartArray[*p3 - 1].outputList.begin(); p2 != m_PartArray[*p3 - 1].outputList.end(); ++p2)
	Downstream_Cap += m_PartArray[p2->id - 1].part.m_c;

      Downstream_Cap += (m_PartArray[*p3 - 1].part.m_cFET * m_PartArray[*p3 - 1].part.m_size);
      m_PartArray[*p3 - 1].part.m_C = Downstream_Cap;
    }
    else {

      if (strcmp(String, "Input")  != 0 &&
	  strcmp(String, "Output") != 0) {

	if (strcmp(String, "pri") == 0)
	  m_PartArray[*p3 - 1].part.m_C = 0.0;
	else
	  m_PartArray[*p3 - 1].part.m_C = (m_PartArray[*p3 - 1].part.m_cFET * m_PartArray[*p3 - 1].part.m_size);

	for (p2 = m_PartArray[*p3 - 1].outputList.begin(); p2 != m_PartArray[*p3 - 1].outputList.end(); ++p2)
	  m_PartArray[*p3 - 1].part.m_C += (m_PartArray[p2->id - 1].part.m_c * m_PartArray[p2->id - 1].part.m_size);
      }
    }
  }
}

void CNetlist::SolveLRSu()
{
  list<PartInfo>::iterator p2;
  vector<int>::iterator p3;
  vector<PartType> tempPartArray;
  char String[10];
  double Old_Cost = 0.0;
  double New_Cost = 0.0;
  int itr = 1;

  for (p3 = m_TO.begin(); p3 != m_TO.end(); ++p3) {
    strcpy(String, m_PartArray[*p3 - 1].part.m_name);
    if (strcmp(String, "pri")    != 0 &&
	strcmp(String, "Input")  != 0 &&
	strcmp(String, "Output") != 0) {
      
      m_PartArray[*p3 - 1].part.m_R = 0.0;
      
      for (p2 = m_PartArray[*p3 - 1].inputList.begin(); p2 != m_PartArray[*p3 - 1].inputList.end(); ++p2) {
	strcpy(String, m_PartArray[p2->id - 1].part.m_name);
	
	if (strcmp(String, "pri")   != 0 &&
	    strcmp(String, "Input") != 0) {  
	  m_PartArray[*p3 - 1].part.m_R += (m_PartArray[p2->id - 1].part.m_mu * (m_PartArray[p2->id - 1].part.m_r / (1.0 * m_PartArray[p2->id - 1].part.m_size)));
	}
	else if (strcmp(String, "pri") == 0)
	  m_PartArray[*p3 - 1].part.m_R += (m_PartArray[p2->id - 1].part.m_mu * m_PartArray[p2->id - 1].part.m_r);
      }

      double op1 = sqrt( (m_PartArray[*p3 - 1].part.m_mu * m_PartArray[*p3 - 1].part.m_r * (m_PartArray[*p3 - 1].part.m_C - (m_PartArray[*p3 - 1].part.m_size * m_PartArray[*p3 - 1].part.m_cFET))) / (1.0 * m_PartArray[*p3 - 1].part.m_c * m_PartArray[*p3 - 1].part.m_R) );
      double op2 = max(m_PartArray[*p3 - 1].part.m_sizeBound[0], op1);
      m_PartArray[*p3 - 1].part.m_size = min(m_PartArray[*p3 - 1].part.m_sizeBound[1], op2);
    }
  }

  for (p3 = m_RTO.begin(); p3 != m_RTO.end(); ++p3) {
    strcpy(String, m_PartArray[*p3 - 1].part.m_name);

    if (m_PartArray[*p3 - 1].part.m_primaryOutput == true) {
      
      double Downstream_Cap = 0.0;
      for (p2 = m_PartArray[*p3 - 1].outputList.begin(); p2 != m_PartArray[*p3 - 1].outputList.end(); ++p2)
	Downstream_Cap += m_PartArray[p2->id - 1].part.m_c;
      
      Downstream_Cap += (m_PartArray[*p3 - 1].part.m_cFET * m_PartArray[*p3 - 1].part.m_size);
      m_PartArray[*p3 - 1].part.m_C = Downstream_Cap; 
    }

    else if (strcmp(String, "Input")  != 0 &&
	     strcmp(String, "Output") != 0) {
      
      if (strcmp(String, "pri") == 0)
	m_PartArray[*p3 - 1].part.m_C = 0.0;
      else
	m_PartArray[*p3 - 1].part.m_C = (m_PartArray[*p3 - 1].part.m_cFET * m_PartArray[*p3 - 1].part.m_size);
      
      for (p2 = m_PartArray[*p3 - 1].outputList.begin(); p2 != m_PartArray[*p3 - 1].outputList.end(); ++p2)
	m_PartArray[*p3 - 1].part.m_C += (m_PartArray[p2->id - 1].part.m_c * m_PartArray[p2->id - 1].part.m_size);
    }
  }

  m_Total_Cost = Old_Cost = ComputeLu(m_PartArray);

  tempPartArray = m_PartArray;

  while( 1 ) {

    for (p3 = m_TO.begin(); p3 != m_TO.end(); ++p3) {
      strcpy(String, tempPartArray[*p3 - 1].part.m_name);
      if (strcmp(String, "pri")    != 0 &&
	  strcmp(String, "Input")  != 0 &&
	  strcmp(String, "Output") != 0) {
	
	tempPartArray[*p3 - 1].part.m_R = 0.0;
	
	for (p2 = tempPartArray[*p3 - 1].inputList.begin(); p2 != tempPartArray[*p3 - 1].inputList.end(); ++p2) {
	  strcpy(String, tempPartArray[p2->id - 1].part.m_name);
	  
	  if (strcmp(String, "pri")   != 0 &&
	      strcmp(String, "Input") != 0) {
	    tempPartArray[*p3 - 1].part.m_R += (tempPartArray[p2->id - 1].part.m_mu * (tempPartArray[p2->id - 1].part.m_r / (1.0 * tempPartArray[p2->id - 1].part.m_size)));
	  }
	  else if (strcmp(String, "pri") == 0)
	    tempPartArray[*p3 - 1].part.m_R += (tempPartArray[p2->id - 1].part.m_mu * tempPartArray[p2->id - 1].part.m_r);
	}

	double op1 = sqrt( (tempPartArray[*p3 - 1].part.m_mu * tempPartArray[*p3 - 1].part.m_r * (tempPartArray[*p3 - 1].part.m_C - (tempPartArray[*p3 - 1].part.m_size * tempPartArray[*p3 - 1].part.m_cFET))) / (1.0 * tempPartArray[*p3 - 1].part.m_c * tempPartArray[*p3 - 1].part.m_R) );
	double op2 = max(tempPartArray[*p3 - 1].part.m_sizeBound[0], op1);
	tempPartArray[*p3 - 1].part.m_size = min(tempPartArray[*p3 - 1].part.m_sizeBound[1], op2);
      }
    }

    for (p3 = m_RTO.begin(); p3 != m_RTO.end(); ++p3) {
      strcpy(String, tempPartArray[*p3 - 1].part.m_name);

      if (tempPartArray[*p3 - 1].part.m_primaryOutput == true) {
	
	double Downstream_Cap = 0.0;
	for (p2 = tempPartArray[*p3 - 1].outputList.begin(); p2 != tempPartArray[*p3 - 1].outputList.end(); ++p2)
	  Downstream_Cap += tempPartArray[p2->id - 1].part.m_c;

	Downstream_Cap += (tempPartArray[*p3 - 1].part.m_cFET * tempPartArray[*p3 - 1].part.m_size);
	tempPartArray[*p3 - 1].part.m_C = Downstream_Cap; 
      }

      else if (strcmp(String, "Input")  != 0 &&
	       strcmp(String, "Output") != 0) {

	if (strcmp(String, "pri") == 0)
	  tempPartArray[*p3 - 1].part.m_C = 0.0;
	else
	  tempPartArray[*p3 - 1].part.m_C = (tempPartArray[*p3 - 1].part.m_cFET * tempPartArray[*p3 - 1].part.m_size);

	for (p2 = tempPartArray[*p3 - 1].outputList.begin(); p2 != tempPartArray[*p3 - 1].outputList.end(); ++p2)
	  tempPartArray[*p3 - 1].part.m_C += (tempPartArray[p2->id - 1].part.m_c * tempPartArray[p2->id - 1].part.m_size);
      }
    }

    New_Cost = ComputeLu(tempPartArray);

    double sum = New_Cost + Old_Cost;
    if (fabs(New_Cost/(1.0 * sum) - Old_Cost/(1.0 * sum)) > m_Condition[0]) {
      m_PartArray  = tempPartArray;
      Old_Cost     = New_Cost;
      m_Total_Cost = New_Cost;
    }
    else break;

  }
}

void CNetlist::SolveLDP()
{
  int    itr = 1;
  double sum;

  ofstream Out;
  Out.open("convergence.txt", ios::out);

  Out << "Lower Bound" << '\t' << "Upper Bound" << endl;
  Out << "===========" << '\t' << "===========" << endl;

  IntializeLM();
  SolveLRSuHelper();

  m_Total_Cost = ComputeLu(m_PartArray);
  ComputeArrivalTime();
  UpdateLM();
  Out << m_Total_Cost << '\t' << m_Total_Delay << endl;
  cout << ".";

  do {
    SolveLRSu();
    ComputeArrivalTime();
    UpdateLM();

    cout << ".";
    itr++;
    Out << m_Total_Cost << '\t' << m_Total_Delay << endl;

    sum = m_Total_Delay + m_Total_Cost;
  }
  while (fabs(m_Total_Delay/(1.0 * sum) - m_Total_Cost/(1.0 * sum)) > m_Condition[1]);

  SolveLRSu();
  ComputeArrivalTime();
  UpdateLM();
  cout << ".";
  Out << m_Total_Cost << '\t' << m_Total_Delay << endl;

  cout << " Done!\nTotal number of iteration: " << itr << endl << endl;

  Out.close();
}

void CNetlist::SGWS_LR()
{
  SolveLDP();
}

void CNetlist::OutputResult(char* filename)
{
  ofstream Out;
  vector<PartType>::iterator p1;

  Out.open(filename, ios::out);
  if ( !Out ) {
    cerr << "File Error: " << filename << " could not be opened" << endl;
    exit( 1 );
  }

  Out << "SGWS_LR ran on " << ctime( &m_time ) << endl;

  SGWS_LR();
  
  m_endTime = clock();
  m_runTime = (double)(m_endTime - m_beginTime) / CLOCKS_PER_SEC;
 
  Out << endl;
  Out << "Statistics" << endl;
  Out << "==========" << endl;
  Out << "Circuit Size     : " << m_nGate << endl;
  Out << "Maximum Delay (s): " << m_Total_Delay << endl;
  Out << "Runtime       (s): " << m_runTime << endl << endl;
  
  Out << "Gate-sizing solution" << endl;
  Out << "====================" << endl;
  Out << "ID\tName\tSize(m^2)"  << endl;
  Out << "--\t----\t---------"  << endl;
  for (p1 = m_PartArray.begin(); p1 != m_PartArray.end(); ++p1) {
    if (strcmp(p1->part.m_name, "pri")    != 0 &&
	strcmp(p1->part.m_name, "Input")  != 0 &&
	strcmp(p1->part.m_name, "Output") != 0) {
      Out << p1->part.m_id << '\t' << p1->part.m_name << '\t' << p1->part.m_size << endl;
    }
  }
  Out.close();
}
