/********************************************************************************
* Fast and Exact Gate Sizing by Lagrangian Relaxation                           *
* File name: Netlist.h; interface for the CNetlist class                        *
* Author: Yu-Hao Wang							        *
* Created      : 09/28/01					  		*
* Last modified: 02/01/02						        *
*********************************************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <list>
#include <algorithm>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "gate.h"
#include "gateType.h"

using namespace std; 

struct PartInfo {            
  int    id,
         pinNum;
  double rt,
         ft;

  PartInfo() {
    id = pinNum = 0;
    rt = ft = 0.0;
  }

  const PartInfo& operator=(const PartInfo &One) {
    if ( &One != this ) {
      id     = One.id;
      pinNum = One.pinNum;
      rt     = One.rt;
      ft     = One.ft;
    }
    return *this;
  }
};

struct PartType {            
  CGate     part;
  list<PartInfo> inputList;
  list<PartInfo> outputList;

  PartType() {
    inputList.clear();	
    outputList.clear();
  }

  const PartType& operator=(const PartType &One) {
    if ( &One != this ) {
      part       = One.part;
      inputList  = One.inputList;
      outputList = One.outputList;
    }
    return *this;
  }

  bool operator==(const PartType &One) const  
  {
    return part.m_id == One.part.m_id;
  }

  bool operator<(const PartType &One) const  
  {
    return part.m_id < One.part.m_id;
  } 
};

class CNetlist  
{
public:

  CNetlist();

  virtual ~CNetlist();

  unsigned length() const { return m_size; }
  bool is_empty() const { return m_size == 0; }

  double maximum(vector<double>);

  void BuildDB(char*); 

  void InputNetlist(char*);

  void TopologicalSort(vector<PartType>, vector<int>&, bool);

  void ComputeArrivalTime();

  void IntializeLM();

  void UpdateLM();

  double ComputeLu(vector<PartType>&);

  void SolveLRSuHelper();

  void SolveLRSu();

  void SolveLDP();

  void SGWS_LR();

  void OutputResult(char*);
  
  double            m_Condition[2],
                    m_driverR,
                    m_loadC,
                    m_Total_Cost,
                    m_Total_Delay;
 private:
  int               m_size,
                    m_nGate,
                    m_nWire,
                    m_nPris,
                    m_nPos,
                    m_nLine;
  time_t            m_time;
  clock_t           m_beginTime,
                    m_endTime;
  double            m_runTime;
  vector<int>       m_TO,
                    m_RTO;
  vector<PartType>  m_PartArray;
  list<CGateType>   m_GateList;
  
};


